@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2010 Edgewall Software
+# Copyright (C) 2011 Edgewall Software
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -19,10 +19,13 @@ we also modify the standard `distutils.command.build` and
 for compiling catalogs are issued upon install.
 """
 
+from __future__ import with_statement
+
 from StringIO import StringIO
 from itertools import izip
 import os
 import re
+from tokenize import generate_tokens, COMMENT, NAME, OP, STRING
 
 from distutils import log
 from distutils.cmd import Command
@@ -37,10 +40,179 @@ try:
                                         compile_catalog, update_catalog
     from babel.messages.pofile import read_po
     from babel.support import Translations
+    from babel.util import parse_encoding
 
     _GENSHI_MARKUP_SEARCH = re.compile(r'\[[0-9]+:').search
 
 
+    _DEFAULT_KWARGS_MAPS = {
+        'Option': {'doc': 4},
+        'BoolOption': {'doc': 4},
+        'IntOption': {'doc': 4},
+        'FloatOption': {'doc': 4},
+        'ListOption': {'doc': 6},
+        'ChoiceOption': {'doc': 4},
+        'PathOption': {'doc': 4},
+        'ExtensionOption': {'doc': 5},
+        'OrderedExtensionsOption': {'doc': 6},
+    }
+
+    _DEFAULT_CLEANDOC_KEYWORDS = (
+        'ConfigSection', 'Option', 'BoolOption', 'IntOption', 'FloatOption',
+        'ListOption', 'ChoiceOption', 'PathOption', 'ExtensionOption',
+        'OrderedExtensionsOption', 'cleandoc_',
+    )
+
+    def extract_python(fileobj, keywords, comment_tags, options):
+        """Extract messages from Python source code, This is patched
+        extract_python from Babel to support keyword argument mapping.
+
+        `kwargs_maps` option: names of keyword arguments will be mapping to
+        index of messages array.
+
+        `cleandoc_keywords` option: a list of keywords to clean up the
+        extracted messages with `cleandoc`.
+        """
+        from trac.util.compat import cleandoc
+
+        funcname = lineno = message_lineno = None
+        kwargs_maps = func_kwargs_map = None
+        call_stack = -1
+        buf = []
+        messages = []
+        messages_kwargs = {}
+        translator_comments = []
+        in_def = in_translator_comments = False
+        comment_tag = None
+
+        encoding = str(parse_encoding(fileobj) or
+                       options.get('encoding', 'iso-8859-1'))
+        kwargs_maps = _DEFAULT_KWARGS_MAPS.copy()
+        if 'kwargs_maps' in options:
+            kwargs_maps.update(options['kwargs_maps'])
+        cleandoc_keywords = set(_DEFAULT_CLEANDOC_KEYWORDS)
+        if 'cleandoc_keywords' in options:
+            cleandoc_keywords.update(options['cleandoc_keywords'])
+
+        tokens = generate_tokens(fileobj.readline)
+        tok = value = None
+        for _ in tokens:
+            prev_tok, prev_value = tok, value
+            tok, value, (lineno, _), _, _ = _
+            if call_stack == -1 and tok == NAME and value in ('def', 'class'):
+                in_def = True
+            elif tok == OP and value == '(':
+                if in_def:
+                    # Avoid false positives for declarations such as:
+                    # def gettext(arg='message'):
+                    in_def = False
+                    continue
+                if funcname:
+                    message_lineno = lineno
+                    call_stack += 1
+                kwarg_name = None
+            elif in_def and tok == OP and value == ':':
+                # End of a class definition without parens
+                in_def = False
+                continue
+            elif call_stack == -1 and tok == COMMENT:
+                # Strip the comment token from the line
+                value = value.decode(encoding)[1:].strip()
+                if in_translator_comments and \
+                        translator_comments[-1][0] == lineno - 1:
+                    # We're already inside a translator comment, continue
+                    # appending
+                    translator_comments.append((lineno, value))
+                    continue
+                # If execution reaches this point, let's see if comment line
+                # starts with one of the comment tags
+                for comment_tag in comment_tags:
+                    if value.startswith(comment_tag):
+                        in_translator_comments = True
+                        translator_comments.append((lineno, value))
+                        break
+            elif funcname and call_stack == 0:
+                if tok == OP and value == ')':
+                    if buf:
+                        message = ''.join(buf)
+                        if kwarg_name in func_kwargs_map:
+                            messages_kwargs[kwarg_name] = message
+                        else:
+                            messages.append(message)
+                        del buf[:]
+                    else:
+                        messages.append(None)
+
+                    for name, message in messages_kwargs.iteritems():
+                        if name not in func_kwargs_map:
+                            continue
+                        index = func_kwargs_map[name]
+                        while index >= len(messages):
+                            messages.append(None)
+                        messages[index - 1] = message
+
+                    if funcname in cleandoc_keywords:
+                        messages = [m and cleandoc(m) for m in messages]
+                    if len(messages) > 1:
+                        messages = tuple(messages)
+                    else:
+                        messages = messages[0]
+                    # Comments don't apply unless they immediately preceed the
+                    # message
+                    if translator_comments and \
+                            translator_comments[-1][0] < message_lineno - 1:
+                        translator_comments = []
+
+                    yield (message_lineno, funcname, messages,
+                           [comment[1] for comment in translator_comments])
+
+                    funcname = lineno = message_lineno = None
+                    kwarg_name = func_kwargs_map = None
+                    call_stack = -1
+                    messages = []
+                    messages_kwargs = {}
+                    translator_comments = []
+                    in_translator_comments = False
+                elif tok == STRING:
+                    # Unwrap quotes in a safe manner, maintaining the string's
+                    # encoding
+                    # https://sourceforge.net/tracker/?func=detail&atid=355470&
+                    # aid=617979&group_id=5470
+                    value = eval('# coding=%s\n%s' % (encoding, value),
+                                 {'__builtins__':{}}, {})
+                    if isinstance(value, str):
+                        value = value.decode(encoding)
+                    buf.append(value)
+                elif tok == OP and value == '=' and prev_tok == NAME:
+                    kwarg_name = prev_value
+                elif tok == OP and value == ',':
+                    if buf:
+                        message = ''.join(buf)
+                        if kwarg_name in func_kwargs_map:
+                            messages_kwargs[kwarg_name] = message
+                        else:
+                            messages.append(message)
+                        del buf[:]
+                    else:
+                        messages.append(None)
+                    kwarg_name = None
+                    if translator_comments:
+                        # We have translator comments, and since we're on a
+                        # comma(,) user is allowed to break into a new line
+                        # Let's increase the last comment's lineno in order
+                        # for the comment to still be a valid one
+                        old_lineno, old_comment = translator_comments.pop()
+                        translator_comments.append((old_lineno+1, old_comment))
+            elif call_stack > 0 and tok == OP and value == ')':
+                call_stack -= 1
+            elif funcname and call_stack == -1:
+                funcname = func_kwargs_map = kwarg_name = None
+            elif tok == NAME and value in keywords:
+                funcname = value
+                func_kwargs_map = kwargs_maps.get(funcname, {})
+                kwarg_name = None
+
+
     def extract_javascript_script(fileobj, keywords, comment_tags, options):
         """Extract messages from Javascript embedding in <script> tags.
 
@@ -135,18 +307,12 @@ try:
                 log.info('generating messages javascript %r to %r',
                          mo_file, js_file)
 
-                infile = open(mo_file, 'rb')
-                try:
+                with open(mo_file, 'rb') as infile:
                     t = Translations(infile, self.domain)
                     catalog = t._catalog
-                finally:
-                    infile.close()
 
-                outfile = open(js_file, 'w')
-                try:
+                with open(js_file, 'w') as outfile:
                     write_js(outfile, catalog, self.domain, locale)
-                finally:
-                    outfile.close()
 
 
     class check_catalog(Command):
@@ -289,17 +455,16 @@ try:
 
 
     def get_command_overriders():
-        # 'bdist_wininst' runs a 'build', so make the latter 
+        # 'bdist_wininst' runs a 'build', so make the latter
         # run a 'compile_catalog' before 'build_py'
         class build(_build):
             sub_commands = [('compile_catalog', None)] + _build.sub_commands
-        
+
         # 'bdist_egg' isn't that nice, all it does is an 'install_lib'
         class install_lib(_install_lib): # playing setuptools' own tricks ;-)
             def l10n_run(self):
-                self.run_command('compile_catalog')                
+                self.run_command('compile_catalog')
             def run(self):
-                from trac.util.compat import any
                 self.l10n_run()
                 # When bdist_egg is called on distribute 0.6.29 and later, the
                 # egg file includes no *.mo and *.js files which are generated
@@ -342,9 +507,38 @@ try:
             'check_catalog_js': check_catalog,
         }
 
+    def get_l10n_trac_cmdclass():
+        build, _install_lib = get_command_overriders()
+        build.sub_commands.insert(0, ('generate_messages_js', None))
+        build.sub_commands.insert(0, ('compile_catalog_js', None))
+        build.sub_commands.insert(0, ('compile_catalog_tracini', None))
+        class install_lib(_install_lib):
+            def l10n_run(self):
+                self.run_command('compile_catalog_tracini')
+                self.run_command('compile_catalog_js')
+                self.run_command('generate_messages_js')
+                self.run_command('compile_catalog')
+        return {
+            'build': build, 'install_lib': install_lib,
+            'check_catalog': check_catalog,
+            'extract_messages_js': extract_messages,
+            'init_catalog_js': init_catalog,
+            'compile_catalog_js': compile_catalog,
+            'update_catalog_js': update_catalog,
+            'generate_messages_js': generate_messages_js,
+            'check_catalog_js': check_catalog,
+            'extract_messages_tracini': extract_messages,
+            'init_catalog_tracini': init_catalog,
+            'compile_catalog_tracini': compile_catalog,
+            'update_catalog_tracini': update_catalog,
+            'check_catalog_tracini': check_catalog,
+        }
+
 
 except ImportError:
     def get_l10n_cmdclass():
         return
     def get_l10n_js_cmdclass():
         return
+    def get_l10n_trac_cmdclass():
+        return
